######################################################################
######################################################################

#
# cport:
#
# create a canvas viewport item in a toplevel window.
#

proc cport {cell top colormap visual args} {
 
  catch {destroy $top};

  toplevel $top	-bd 0 \
		-colormap $colormap \
		-visual $visual;
  wm geometry $top 400x340;

  canvas $top.c;
  pack $top.c	-expand 1 \
		-fill both \
		-padx 10 \
		-pady 10;

  $top.c configure -bg steelblue;
  
  set id [$top.c create viewport 30 20 350 260 \
		-minu -100 \
		-minv -100 \
		-maxu 100 \
		-maxv 100 \
		-fill black \
		-outline red \
		-projection perspective \
		-vrp {0 0 100} \
		-prp {0 0 5} \
		-vpn {0.1 0 1} \
		-vup {0 1 0} \
		-ambient #5fff5fff5fff \
		-depthcue {{0 0 0} linear 0 1000} \
		-light0 {{1 1 1} {1 1 1} {1 1 1} {50 50 50 0}} \
		-light1 {{1 1 1} {1 1 1} {1 1 1} {-50 50 50 0}} \
		];

  $top.c itemconfigure $id -cell $cell;
  if {$args != ""} {eval [concat "{$top.c} itemconfigure $id" $args]};
  update;
}


######################################################################
######################################################################

#
# vport:
#
# create a viewport widget in a toplevel window.
#

proc vport {cell top colormap visual args} {

  catch {destroy $top};

  toplevel $top	-relief sunken \
		-bd 0 \
		-colormap $colormap \
		-visual $visual;
  wm geometry $top 400x320;

  viewport $top.c 1 1;
  pack $top.c	-expand 1 \
		-fill both \
		-padx 10 \
		-pady 10;
  
  $top.c configure	-cell $cell \
			-minu -100 \
			-minv -100 \
			-maxu 100 \
			-maxv 100 \
			-fill black \
			-outline red \
			-width 1 \
			-projection perspective \
			-vrp {0 0 100} \
			-prp {0 0 5} \
			-vpn {0.1 0 1} \
			-vup {0 1 0} \
			-ambient #5fff5fff5fff \
			-depthcue {{0 0 0} linear 0 1000} \
			-light0 {{1 1 1} {1 1 1} {1 1 1} {50 50 50 0}} \
			-light1 {{1 1 1} {1 1 1} {1 1 1} {-50 50 50 0}};

  if {$args != ""} {eval [concat "$top.c config " $args]};
  update;

  bind $top.c <Button-1> {start_navigation %W {100 100 100} None 0.2};
  bind $top.c <ButtonRelease-1> {stop_navigation};

  bind $top.c <Control-Button-1> {start_navigation %W {100 100 100} Control 0.2};
  bind $top.c <ButtonRelease-1> {stop_navigation};

  bind $top.c <Shift-Button-1> {start_navigation %W {100 100 100} Shift 0.2};
  bind $top.c <ButtonRelease-1> {stop_navigation};

  bind $top.c <Button-2> {pick_model %W %x %y};
}

######################################################################
######################################################################

#
# primitives:
#
# populate a cell with primitive objects:
#
# polygon, cylinder, disk, sphere
#

proc primitives {cell} {

  if ![catch {cell create $cell}] {

    # type point

    $cell create point point1	-vertices [list {-10 -10 -10} \
					        { 10 -10 -10} \
					        { 10 -10  10} \
					        {-10 -10  10} \
					        {-10  10 -10} \
					        { 10  10 -10} \
					        { 10  10  10} \
					        {-10  10  10} \
					        ] \
				-pos {-75 0 0} \
				-fwd {-0.5 0 1} \
				-up {0 1 0} \
				-pointsize 4 \
				-shading smooth \
				-color green \
				-emissive {0 0 0} \
				-ambient {0.4 0.4 0.4} \
				-diffuse {1 1 1} \
				-specular {1 1 1 15};
    # type sphere

    $cell create sphere sphere1 \
	-radius 10 \
	-slices 20 \
	-stacks 15 \
	-pos {-45 0 0} \
	-fwd {0 0 1} -up {0 1 0} \
	-shading smooth \
	-normals smooth \
	-color red \
	-emissive {0.8 0.8 0.8} \
	-ambient {0 0 0} \
	-diffuse {0.5 0.5 0.5} \
	-specular {1 1 1 20};
	
    # type cylinder

    $cell create cylinder cyl1 \
	-topradius 0 \
	-baseradius 10 \
	-height 20 \
	-slices 20 \
	-stacks 15 \
	-pos {-15 -5 0} \
	-fwd {0 1 0} -up {0 0 1} \
	-shading smooth \
	-normals smooth \
	-color orange \
	-emissive {0 0 0} \
	-ambient {0.4 0.4 0.4} \
	-diffuse {1 1 1} \
	-specular {1 1 1 15};
	
    # type disk

    $cell create disk disk1 \
	-innerradius 5 \
	-outerradius 10 \
	-loops 20 \
	-pos {15 0 0} \
	-fwd {-1 2 2} -up {0 1 0} \
	-shading smooth \
	-normals smooth \
	-color palegreen \
	-emissive {0 0 0} \
	-ambient {0.4 0.4 0.4} \
	-diffuse {1 1 1} \
	-specular {1 1 1 15};

    # type polygon

    $cell create polygon cube1	-vertices [list {-10 -10 -10} \
					        { 10 -10 -10} \
					        { 10 -10  10} \
					        {-10 -10  10} \
					        {-10  10 -10} \
					        { 10  10 -10} \
					        { 10  10  10} \
					        {-10  10  10} \
					        ] \
			      	-surfaces [list {1 2 3 4} \
					        {8 7 6 5} \
					        {1 4 8 5} \
					        {2 6 7 3} \
					        {1 5 6 2} \
					        {3 7 8 4} \
					        ] \
				-pos {45 0 0} \
				-fwd {-0.5 0 1} \
				-up {0 1 0} \
				-shading smooth \
				-color blue \
				-emissive {0 0 0} \
				-ambient {0.4 0.4 0.4} \
				-diffuse {1 1 1} \
				-specular {1 1 1 15} \
				-average off;

    # type line

    $cell create line line1	-vertices [list {-10 -10 -10} \
					        { 10 -10 -10} \
					        { 10 -10  10} \
					        {-10 -10  10} \
					        {-10  10 -10} \
					        { 10  10 -10} \
					        { 10  10  10} \
					        {-10  10  10} \
					        ] \
			      	-lines [list {1 2 3 4} \
					     {8 7 6 5} \
					     {1 4 8 5} \
					     {2 6 7 3} \
					     {1 5 6 2} \
					     {3 7 8 4} \
					     ] \
				-pos {75 0 0} \
				-fwd {-0.5 0 1} \
				-up {0 1 0} \
				-linewidth 4 \
				-linestipple 0xcccc \
				-linestipplefactor 2 \
				-shading smooth \
				-color brown \
				-emissive {0 0 0} \
				-ambient {0.4 0.4 0.4} \
				-diffuse {1 1 1} \
				-specular {1 1 1 15};

  }
  return [$cell list];
}


######################################################################
######################################################################

proc pick_model {port x y} {

  $port configure -pick [list $x $y 1 1];
  puts [lindex [$port cget -pick] 1];

}
